"""MAJIS ITL timeline sub-module."""

import warnings
from collections import defaultdict, namedtuple
from pathlib import Path

from numpy import datetime64
from openpyxl import Workbook, load_workbook
from packaging.version import Version

from planetary_coverage import datetime
from planetary_coverage.events import EventsDict, EventsList
from planetary_coverage.html import table

from ..misc import fmt_datetime
from ..misc.events import flatten, flatten_events
from ..misc.evf import ref_key
from ..misc.time import fmt_timedelta
from .reader import read_itl

DEFAULT_TEMPLATE = Path(__file__).parent / 'timeline.xlsm'

BINNING = {
    '1': 'No Binning',
    '2': 'Binning x2',
    '4': 'Binning x4',
}

ITL_MAPPING = {
    'OBS_NAME': ('OBS_NAME', str),
    'start_angle': ('START_ANGLE', float),  # °
    'start_scan_speed': ('START_SCAN_SPEED', float),  # °/s
    'stop_scan_speed': ('STOP_SCAN_SPEED', float),  # °/s
    'Scanner step per frame': ('SYNCHRONOUS', float),  # (-3, 0, 3)
    'stop_angle': ('STOP_ANGLE', float),  # °
    'First CU_frame start (UTC)': ('t_start', fmt_datetime),  # YYYY-MM-DDThh:mm:ss.msZ
    'Last CU_frame stop (UTC)': ('t_end', fmt_datetime),  # YYYY-MM-DDThh:mm:ss.msZ
    'cu_trep_ms': ('CU_TREP', lambda s: int(str(s).replace('ms', ''))),  # ms
    'spatial_binning': ('BINNING', lambda i: BINNING[str(i)]),
    'nb_cu_frames_tot': ('CU_FRAME', int),
    'ppe': ('PPE', int),
    'Start Row VI': ('START_ROW_VIS', int),
    'prime': ('PRIME', lambda b: 'MAJIS' if b else 'other'),
    'Comments': ('COMMENTS', lambda s: s if s else None),
    'ITL name': ('ITL', lambda p: Path(p).name if p else None),
}

ChangeLog = namedtuple('Log', 'version,date,author,change')


class TimelineChangeLog:
    """MAJIS timeline change log."""

    def __init__(self, template: Workbook, sheet_name='change log'):
        self.log = self.read_changelog(list(template[sheet_name])[1:])

    def __str__(self):
        return '\n\n'.join(
            f'>>> {str(log.version):>6} | {log.date} | {log.author}\n{log.change}'
            for log in self.log
        )

    def __repr__(self):
        return str(self)

    def __getitem__(self, i):
        return self.log[i]

    def __iter__(self):
        return iter(self.log)

    def _repr_html_(self):
        return table(self.log, header=['Version', 'Date', 'Author', 'Changes'])

    def read_changelog(self, rows: list) -> (list, list):
        """Read changelog changes grouped by version, date and author."""
        log = defaultdict(list)

        # Group release changes
        key = None, None, None
        for row in rows:
            version, date, change, author = (cell.value for cell in row[:4])

            if change and change != '…':
                key = (
                    Version(str(version)) if version else key[0],
                    str(date.date()) if date else key[1],
                    str(author) if author else key[2],
                )

                log[key].extend(change.splitlines())

        # Reverse changes order (latest first) and add line returns between changes
        return [ChangeLog(*key, '\n'.join(log[key])) for key in reversed(log)]


class Timeline:
    """MAJIS timeline from template.

    You can either append an existing timeline template
    or use the default one.

    Warning
    -------
    Some extensions (`Data Validation`) are not included
    into the exported file. This is a openpyxl limitation.

    """

    sheet_name = 'Timeline'

    def __init__(
        self,
        observations: str | Path | EventsList | EventsDict | None = None,
        timeline: str | Path | None = None,
        ca_ref: str | dict | None = None,
        refs: dict | str | list | None = None,
    ):
        self.fname = Path(timeline) if timeline else DEFAULT_TEMPLATE
        self._load_data()

        # Set global reference time w.r.t. C/A
        self.ca_ref = ca_ref

        if observations:
            self.append(observations, refs=refs)

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.fname}')"

    def __len__(self) -> int:
        return len(
            [
                obs
                for obs in self._timeline[self.fields['OBS_NAME']][self.header :]
                if obs.value
            ]
        )

    def __getitem__(self, item: str | int | tuple) -> list | dict | str | int | float:
        if isinstance(item, str):
            if item in self.fields:
                return [
                    col.value
                    for col in self._timeline[self.fields[item]][
                        self.header : self.header + len(self)
                    ]
                ]
            raise KeyError(f'Unknown `{item}` key')

        if isinstance(item, int):
            if 1 <= item <= len(self):
                return {
                    key: cell.value
                    for key, cell in zip(
                        self.fields, self._timeline[self.header + item], strict=True
                    )
                    if cell.value and not str(cell.value).startswith('=')
                }
            raise IndexError(f'Invalid index: {item} not between 1 and {len(self)}')

        if isinstance(item, tuple):
            name, i = item
            return self[name][i - 1]

        raise TypeError(
            f'Only `str`, `int` or `str, int` are accepted (`{type(item).__name__}` provided)'
        )

    def __setitem__(self, item: (str, int), value: str | int | float) -> None:
        name, i = item
        self._timeline[self.fields[name] + str(i + self.header)] = value

    def _repr_html_(self):
        return table(
            [
                [cell.value for cell in row]
                for row in self._timeline[f'{self.header + 1}:{self.header + len(self)}']
            ],
            header=list(self.fields),
        )

    def _load_data(self) -> None:
        """Load timeline data."""
        # Disable openpyxl warning about unsupported extension (`Data validation`)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            self._template = load_workbook(self.fname, keep_vba=True)

        # Load Timeline spreadsheet
        self._timeline = self._template[self.sheet_name]

        # Load Timeline header fields mapping
        self.fields = {cell.value: cell.column_letter for cell in self._timeline[1]}

        # Load template changelog
        if 'template change log' in self._template.sheetnames:
            sheet_name = 'template change log'
        else:
            sheet_name = 'change log'

        self.log = TimelineChangeLog(self._template, sheet_name=sheet_name)

    @property
    def version(self):
        """Template version from changelog."""
        return self.log[0].version

    @property
    def header(self):
        """Timeline header size."""
        return 2 if self.version < Version('2.0') else 3

    def append(
        self,
        observations: str | Path | EventsList | EventsDict,
        refs: dict | str | list | None = None,
    ) -> None:
        """Append ITL observations blocks."""
        if isinstance(observations, str | Path):
            observations = read_itl(observations, refs=refs)

        for i, obs in enumerate(flatten(observations), start=len(self) + 1):
            # Required fields
            for field, (key, fmt) in ITL_MAPPING.items():
                self[field, i] = fmt(obs[key])

            # Additional fields
            self['Mirror Flag', i] = (
                'ENABLE'
                if float(obs['START_SCAN_SPEED']) or float(obs['STOP_SCAN_SPEED'])
                else 'DISABLE'
            )

            self['First CU_frame start wrt C/A', i] = self._fmt_timedelta(obs.start)
            self['Last CU_frame stop wrt C/A', i] = self._fmt_timedelta(obs.stop)

    @property
    def ca_ref(self) -> datetime:
        """C/A reference for relative time."""
        return self._ca_ref

    @ca_ref.setter
    def ca_ref(self, ca_ref):
        """Set relative time w.r.t. C/A reference.

        Warning
        -------
        All the previous relative values will be recomputed if this value is changed.

        """
        if ca_ref:
            _, self._ca_ref = ref_key(ca_ref)

            # Recompute relative values for all observation w.r.t. C/A reference
            for i in range(1, len(self) + 1):
                self['First CU_frame start wrt C/A', i] = self._fmt_timedelta(
                    self['First CU_frame start (UTC)', i]
                )
                self['Last CU_frame stop wrt C/A', i] = self._fmt_timedelta(
                    self['Last CU_frame stop (UTC)', i]
                )
        else:
            self._ca_ref = None

    def _fmt_timedelta(self, t: str | datetime64) -> str | None:
        """Compute relative time w.r.t. C/A reference."""
        return fmt_timedelta(datetime(t) - self.ca_ref) if self.ca_ref else None

    def save(self, fout: str | Path | None = None) -> Path:
        """Save MAJIS timeline."""
        if fout:
            fout = Path(fout)
        elif self.fname != DEFAULT_TEMPLATE:
            fout = self.fname
        else:
            raise FileExistsError(
                'Can not overwrite default template. Please provide and export filename.'
            )

        self._template.save(fout)

        return fout

    @property
    def science(self) -> TimelineChangeLog:
        """Science change log.

        Raises
        ------
        ValueError:
            Only available for template ≥ 2.0

        """
        if self.version < Version('2.0'):
            raise ValueError(
                f'Science changelog is only available in template ≥ 2.0 (current: `{self.version}`)'
            )

        return TimelineChangeLog(self._template, sheet_name='science change log')


@flatten_events
def save_xlsm(
    fout: str | Path | None,
    *events: EventsList | EventsDict,
    ca_ref: str | dict | None = None,
    overlap: bool = False,
    timeline: str | Path | None = None,
) -> Path:
    """Save ITL events to XLSM timeline.

    If a timeline is provided but no explicit output file
    the output file will be same in the original timeline.

    """
    return Timeline(events, timeline=timeline, ca_ref=ca_ref).save(fout)



"""ITL schema sub-module."""

import re
from collections import UserDict
from json import loads
from pathlib import Path

from jsonschema import validate

SCHEMA_FOLDER = Path(__file__).parent / 'schema'


class ItlJsonSchema(UserDict):
    """ITL JSON schema object."""

    def __init__(self, fname: str | Path):
        self.fname = Path(fname)
        self.data = loads(self.fname.read_text(encoding='utf-8'))

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}> {self}'

    @property
    def name(self) -> str:
        """Schema file name."""
        return self.fname.stem

    @property
    def title(self) -> str:
        """Schema title description."""
        return self.data.get('title', '(undefined)')

    @property
    def version(self) -> int:
        """Schema version.

        At the moment this value is extracted from the title property.

        """
        if 'version' in self.data:
            return int(self.data['version'])

        for version in re.findall(r'v(\d+)', self.title):
            return int(version)

        return 0

    def validate(self, itl: dict) -> None:
        """Validate ITL JSON content."""
        validate(instance=itl, schema=self.data)


MAJIS_ITL_SCHEMAS = [
    ItlJsonSchema(SCHEMA_FOLDER / 'jsoc-itl-schema.json'),
    ItlJsonSchema(SCHEMA_FOLDER / 'majis-itl-schema.json'),
]



"""Events module."""

from operator import attrgetter

from planetary_coverage.events.event import (
    AbstractEvent,
    EventsDict,
    EventsList,
)


def flatten(events: tuple | list | EventsList | EventsDict | AbstractEvent) -> EventsList:
    """Flatten multi events object."""
    if isinstance(events, tuple | list | EventsList | EventsDict):
        return EventsList(
            sorted(
                [
                    ev
                    for event in events
                    for ev in (
                        flatten(event)
                        if isinstance(event, EventsList | EventsDict)
                        else [event]
                    )
                ],
                key=attrgetter('start'),
            )
        )

    if isinstance(events, AbstractEvent):
        return EventsList([events])

    raise TypeError(f'Invalid events type: `{type(events)}`')


def group(events: tuple | list | EventsList | EventsDict) -> EventsDict:
    """Group events by observation name."""
    if isinstance(events, tuple | list | EventsList):
        return EventsDict(events)

    if isinstance(events, EventsDict):
        return events

    raise TypeError(f'Invalid events type: `{type(events)}`')


def concatenate(
    *events: EventsList | EventsDict, flat: bool = False, overlap: bool = False
) -> EventsList | EventsDict:
    """Concatenate ITL events.

    Note
    ----
    By default, concatenated blocks must not overlap
    each other. This can be disable with `overlap=True`.

    """
    blocks = flatten(events)

    if blocks and not overlap:
        previous = blocks[0]
        for block in blocks[1:]:
            if block.start < previous.stop:
                raise ValueError(f'Overlap between `{previous}` and `{block}`')
            previous = block

    return blocks if flat else group(blocks)


def flatten_events(func):
    """Events flatten decorator.

    Mainly used as a prolog before saving files.

    """

    def wrapper(
        fout: ...,
        *events: EventsList | EventsDict,
        overlap: bool = False,
        **kwargs,
    ):
        """Concatenate events blocks before the calling the main function."""
        blocks = concatenate(*events, flat=True, overlap=overlap)
        return func(fout, *blocks, **kwargs)

    return wrapper



"""ITL JSON mapping with EPS format."""

from ...misc import fmt_datetime

EPS_MAPPING_OBS = {
    'OBS_KEY': ('name', str),
    't_start': ('start_time', fmt_datetime),
    't_end': ('end_time', fmt_datetime),
    'TYPE': ('type', str),
    'OBSERVATION_TYPE': ('observation_type', str),
    'INSTRUMENT': ('instrument', str),
    'OBS_NAME': ('unique_id', str),
    'TARGET': ('target', str),
    'COMMENTS': ('comment', str),
}

EPS_MAPPING_PARAMS = {
    'SCENARIO': ('scenario_id', str),
    'CU_TREP': ('cu_trep_ms', lambda v: int(str(v).replace('ms', ''))),
    'CU_FRAME': ('nb_cu_frames_tot', int),
    'BINNING': ('spatial_binning', int),
    'PPE': ('ppe', int),
    'START_ROW_VIS': ('start_row_vi', int),
    'START_ANGLE': ('start_angle', float),
    'STOP_ANGLE': ('stop_angle', float),
    'SYNCHRONOUS': ('scanner_step_per_frame', int),
    'START_SCAN_SPEED': ('start_scan_speed', float),
    'STOP_SCAN_SPEED': ('stop_scan_speed', float),
}

JSON_MAPPING = {
    json: eps for eps, (json, _) in (EPS_MAPPING_OBS | EPS_MAPPING_PARAMS).items()
}



"""ITL JSON export module."""

import getpass
from datetime import datetime
from pathlib import Path

from planetary_coverage.events import EventsDict, EventsList

from ...misc import fmt_datetime
from ...misc.events import flatten_events
from ...misc.export import save_file
from .mapping import EPS_MAPPING_OBS, EPS_MAPPING_PARAMS
from .schema import MAJIS_ITL_SCHEMAS


@flatten_events
def save_itl_json(
    fout: str | Path | None,
    *events: EventsList | EventsDict,
    ref: str | None = None,
    overlap: bool = False,
) -> Path | dict:
    """Save ITL events to a new JSON ITL file.

    Notes
    -----
    By default, ITL blocks must not overlap each other.
    This can be disable with `overlap=True`.

    """
    if ref is not None:
        raise NotImplementedError

    content = {
        'header': get_header(fout),
        'timeline': [fmt_json(event) for event in events],
    }

    for schema in MAJIS_ITL_SCHEMAS:
        schema.validate(content)

    return save_file(fout, content, suffix='.json')


def get_header(fout: str | Path | None = None) -> dict:
    """Get JSON header."""
    return {
        'filename': 'ITL_.json' if fout is None else Path(fout).name,
        'creation_date': fmt_datetime(datetime.now()),
        'author': getpass.getuser(),
    }


EXCLUDED_KEYS = [
    'ITL',
    'PRIME',
]


def fmt_json(event: EventsList | EventsDict) -> dict:
    """Format event to JSON timeline."""
    content, params = {'name': event.key}, {}
    for eps_key, value in event.items():
        if eps_key in EXCLUDED_KEYS:
            continue

        if eps_key in EPS_MAPPING_OBS:
            json_key, fmt = EPS_MAPPING_OBS[eps_key]
            content[json_key] = fmt(value)

        elif eps_key in EPS_MAPPING_PARAMS:
            json_key, fmt = EPS_MAPPING_PARAMS[eps_key]
            params[json_key] = fmt(value)

        else:
            params[eps_key.lower()] = value

    # Add observation type if missing
    if 'type' not in content:
        content['type'] = 'OBSERVATION'

    # Append parameters after the other obs parameters
    content['parameters'] = params

    # Put comment at the end of the block (if present)
    if 'comment' in content:
        content['comment'] = content.pop('comment')

    return content



"""ITL generic reader module."""

from pathlib import Path

from planetary_coverage.events import EventsDict, EventsList

from ..misc.events import flatten, group
from .eps import read_itl_eps
from .json import read_itl_json


def read_itl(
    fname: str | Path,
    refs: dict | str | list | None = None,
    flat: bool = False,
) -> EventsDict | EventsList:
    """Read any ITL file.

    Both, EPS/text (``.itl`` or ``.txt``) or JSON (``.json``)
    format are supported.

    In the case of EPS format, relative datetime input is supported
    with references value(s) or file (``.evf``).

    The result can be return as a nested dict of events (default) or
    as a flat list of events.

    """
    f = Path(fname)

    match f.suffix.lower():
        case '.itl' | '.txt':
            events = read_itl_eps(f, refs=refs)
        case '.json':
            if refs:
                raise NotImplementedError('Relative time ITL is not yet supported.')
            events = read_itl_json(f)
        case _:
            raise ValueError(f'Invalid ITL file input: {fname}')

    return flatten(events) if flat else group(events)



"""ITL EPS export module."""

from pathlib import Path

from planetary_coverage.events import EventsDict, EventsList, EventWindow

from ...misc import fmt_datetime
from ...misc.events import flatten_events
from ...misc.evf import ref_key
from ...misc.export import save_file


@flatten_events
def save_itl_eps(
    fout: str | Path | None,
    *events: EventsList | EventsDict,
    ref: str | None = None,
    overlap: bool = False,
    header: str = None,
) -> Path:
    """Save ITL events to a new EPS ITL file.

    Notes
    -----
    By default, ITL blocks must not overlap each other.
    This can be disable with `overlap=True`.

    Any custom header can be added.

    Reference datetime, if provided, will be reported after
    the header (if any provided).

    All the bocks will be commented if additional parameters
    are present in the Event objects.

    """
    content = [header, ''] if header else []

    if ref:
        ref = ref_key(ref)

        content += [
            '# Relative time reference:',
            f'# {ref[1]}  {ref[0]}',
            '',
        ]

    for event in events:
        content.extend(fmt_eps_comments(event))

        start, end = fmt_datetime(event.start, event.stop, ref=ref)

        inst = event['INSTRUMENT']
        prime = ' (PRIME=TRUE)' if event['PRIME'] else ''

        content.append(f'{start}  {inst}  OBS_START  {event.key}{prime}')
        content.append(f'{end}  {inst}  OBS_END    {event.key}')
        content.append('')  # empty line

    return save_file(fout, content, suffix='.itl')


EXCLUDED_KEYS = [
    'INSTRUMENT',
    'PRIME',
    'T_START',
    'T_END',
    'COMMENTS',
    'ITL',
]


def fmt_eps_comments(event: dict | EventWindow, max_length: int = 90) -> list[str]:
    """Format EPS comments."""
    if instrument := event.get('INSTRUMENT'):
        prefix = f'# {instrument.upper()} -'
    else:
        prefix = '#'

    comments, current = [], prefix
    for key, value in event.items():
        if key.upper() in EXCLUDED_KEYS:
            continue

        param = f'{key.upper()}={value}'

        # Append only if length is shorter than `max_length`
        if current == prefix or len(current) + len(param) <= max_length:
            current += f' {param}'
        else:
            comments.append(current)
            current = f'{prefix} {param}'

    # Add last parameter key-value is not empty
    if current != prefix:
        comments.append(current)

    # Add multi-lines comments (separated with a `/`)
    if 'COMMENTS' in event and event['COMMENTS']:
        for comment in event['COMMENTS'].split('/'):
            comments.append(f'{prefix} COMMENT: {comment.strip()}')

    return comments



"""MAJIS ITL submodule."""

from .export import save_csv, save_itl, save_xlsm
from .reader import read_itl
from .timeline import Timeline

__all__ = [
    'read_itl',
    'save_itl',
    'save_csv',
    'save_xlsm',
    'Timeline',
]



"""ITL export module."""

from pathlib import Path

from planetary_coverage.events import EventsDict, EventsList

from .csv import save_csv
from .eps import save_itl_eps
from .json import save_itl_json
from .timeline import save_xlsm


def save_itl(
    fout: str | Path | None,
    *events: EventsList | EventsDict,
    ref: str | dict | None = None,
    overlap: bool = False,
    **kwargs,
) -> Path:
    """Save ITL events to a new ITL file.

    Note
    ----
    By default, ITL blocks must not overlap each other.
    This can be disable with ``overlap=True``.

    """
    ext = Path(fout).suffix

    match ext:
        case '.itl':
            return save_itl_eps(
                fout,
                *events,
                ref=ref,
                overlap=overlap,
                **kwargs,
            )
        case '.json':
            return save_itl_json(
                fout,
                *events,
                ref=ref,
                overlap=overlap,
                **kwargs,
            )
        case '.csv':
            return save_csv(
                fout,
                *events,
                ref=ref,
                overlap=overlap,
                **kwargs,
            )
        case '.xls' | '.xlsx' | '.xlsm':
            return save_xlsm(
                fout,
                *events,
                ca_ref=ref,
                overlap=overlap,
                **kwargs,
            )
        case _:
            raise NotImplementedError



"""Majis ITL in EPS format sub-module."""

from .export import save_itl_json
from .reader import read_itl_json
from .schema import MAJIS_ITL_SCHEMAS

__all__ = [
    'MAJIS_ITL_SCHEMAS',
    'read_itl_json',
    'save_itl_json',
]



"""Majis ITL in EPS format sub-module."""

import re
from pathlib import Path

from numpy import datetime64

from planetary_coverage.events import EventWindow

from ...misc import get_datetime

# ITL prefix pattern: # INST - KEY=VALUE ... or # INST - COMMENTS : VALUE
ATTRIBUTES = re.compile(r'^#\s*\w+\s+-\s+(?P<values>.*)')
COMMENTS = re.compile(r'^#\s*\w+\s+-\s+COMMENTS?\s*:\s*(?P<value>.*)')
OBS_INST = re.compile(r'(?P<inst>[a-zA-Z]\w*)\s+OBS_(?:START|END)\s')
OBS_KEY = re.compile(r'OBS_(?:START|END)\s+(?P<key>\w*[a-zA-Z])(?:_\d+)?')


def read_itl_eps(
    fname: str | Path,
    refs: dict | str | list | None = None,
) -> list[EventWindow]:
    """Read ITL file in EPS (text) format.

    Note
    ----
    - The blocks can be prefixed with additional instrument parameters.

    - Blocks must be continuous, ie. consecutive OBS_START and OBS_END lines
      should have the same instrument observation name.
      If not an ValueError will be raised

    """
    lines = Path(fname).read_text().splitlines()

    return _parse_itl_eps(lines, refs=refs, filename=fname)


def _parse_itl_eps(
    lines: list[str],
    refs: dict | str | list | None = None,
    filename: str | Path | None = None,
) -> list[EventWindow]:
    """Parse ITL EPS content as EventWindows list."""
    events, attrs, comments, inst, key = [], {}, [], None, None

    for line in lines:
        if line.startswith('#'):
            comments, attrs = _parse_itl_eps_comment(line, comments, attrs)
            continue

        if ' OBS_START ' in line:
            start, inst, key = _parse_itl_eps_obs(line, refs)

            attrs['PRIME'] = '(PRIME=TRUE)' in line
            continue

        if ' OBS_END ' in line:
            end, _inst, _key = _parse_itl_eps_obs(line, refs)

            _check_obs_block(inst, key, _inst, _key)

            attrs['COMMENTS'] = ' / '.join(comments) if comments else None
            attrs['ITL'] = Path(filename) if filename else None

            # Compile the observation in an EventWindow object
            event = EventWindow(key, t_start=start, t_end=end, INSTRUMENT=inst, **attrs)
            events.append(event)

        attrs, comments, inst, key = {}, [], None, None

    return events


def _parse_itl_eps_comment(line: str, comments: list, attrs: dict) -> (list, dict):
    """Parse ITL EPS comment line.

    Extract new comments or attributes if present.

    """
    if match := COMMENTS.match(line):
        comments.append(match.group('value').strip())

    elif match := ATTRIBUTES.match(line):
        kv = [field.split('=', 1) for field in match.group('values').split(' ')]
        attrs |= dict(kv)

    return comments, attrs


def _parse_itl_eps_obs(
    line: str, refs: dict | str | list | None
) -> (datetime64, str, str):
    """Parse ITL EPS observation line."""
    dt = get_datetime(line, refs=refs)

    if not (match := OBS_INST.search(line)):
        raise ValueError(f'Missing instrument in: `{line}`')

    inst = match.group('inst')

    if not (match := OBS_KEY.search(line)):
        raise ValueError(f'Missing obs name in: `{line}`')

    key = match.group('key')

    return dt, inst, key


def _check_obs_block(inst: str, key: str, _inst: str, _key: str) -> bool:
    """Check observation block consistency.

    Raise a ValueError if the instrument or key mismatch in an observation block.

    """
    if _inst != inst:
        raise ValueError(f'Instrument block mismatch: `{inst}` / `{_inst}`')

    if _key != key:
        raise ValueError(f'Obs name block mismatch: `{key}` / `{_key}`')



"""ITL CLI module."""

import json
import sys
from argparse import ArgumentParser
from pathlib import Path

from ..misc import read_evf
from .export import save_csv, save_itl, save_itl_eps, save_itl_json, save_xlsm
from .reader import read_itl


def cli(argv: list | None = None):
    """ITL command line interface."""
    parser = ArgumentParser(description='MAJIS ITL toolbox')
    parser.add_argument(
        'itl',
        help='Input ITL filename(s). If multiple files are provided '
        'they will be concatenated.',
        nargs='*',
        metavar='input.[itl|json]',
    )
    parser.add_argument(
        '-o',
        '--output',
        help='Output filename, it could be either ITL, JSON, CSV or XLSM. '
        'If none provided, the results will be displayed (only for ITL, JSON and CSV).',
        metavar='output.[itl|json|csv|xlsm]',
        default='',
    )
    parser.add_argument(
        '-f',
        '--force',
        action='store_true',
        help='Overwrite the output file if already exists.',
    )
    parser.add_argument(
        '-t',
        '--time-ref',
        help='Input events time reference(s). '
        'If multiple values are required use an `events.evf` file.',
        metavar='"YYYY-MM-DDTHH:MM:SS REF_NAME (COUNT = N)"',
    )
    parser.add_argument(
        '-r',
        '--relative-to',
        help='Reference time to be used for relative time output.',
        metavar='"YYYY-MM-DDTHH:MM:SS REF_NAME (COUNT = N)"',
    )
    parser.add_argument(
        '--timeline',
        help='Original timeline to append. '
        'If no explicit `--output` is provided new observations '
        'will be append in this file.',
        metavar='timeline.xlsm',
    )
    parser.add_argument(
        '--header',
        help='ITL custom file header.',
        metavar='"# my-custom-header"',
    )
    parser.add_argument(
        '--overlap',
        action='store_true',
        help='Allow blocks overlap.',
    )
    parser.add_argument(
        '--json',
        action='store_true',
        help='Display the ITL as JSON.',
    )
    parser.add_argument(
        '--csv',
        action='store_true',
        help='Display the ITL as CSV.',
    )
    parser.add_argument(
        '--csv-sep',
        help='CSV separator (default: ";")',
        default=';',
        metavar='separator',
    )

    args, _ = parser.parse_known_args(argv)

    refs = read_evf(args.time_ref)

    events = [read_itl(itl, refs=refs, flat=True) for itl in args.itl]

    if args.output:
        fout = Path(args.output)

        if not args.force and fout.exists():
            raise FileExistsError(args.output)

        match fout.suffix.lower():
            case '.itl':
                kwargs = {'header': args.header}
            case '.json':
                kwargs = {}
            case '.csv':
                kwargs = {'sep': args.csv_sep}
            case '.xlsm':
                kwargs = {'timeline': args.timeline}
            case _:
                raise ValueError(
                    'Only `.itl`|`.json`|`.csv`|`.xlsm` extension are accepted '
                    f'(not `{fout.suffix}`)'
                )

        save_itl(
            fout,
            *events,
            ref=args.relative_to,
            overlap=args.overlap,
            **kwargs,
        )

        sys.stdout.write(f'ITL saved in: {fout}\n')

    elif args.timeline:
        save_xlsm(
            None,
            *events,
            timeline=args.timeline,
            ca_ref=args.relative_to,
            overlap=args.overlap,
        )
        sys.stdout.write(f'ITL appended to: {args.timeline}\n')

    elif args.json:
        sys.stdout.write(
            json.dumps(
                save_itl_json(None, *events, overlap=args.overlap),
                indent=2,
            )
            + '\n'
        )

    elif args.csv:
        sys.stdout.write(
            '\n'.join(save_csv(None, *events, overlap=args.overlap, sep=args.csv_sep))
            + '\n'
        )

    else:
        sys.stdout.write(
            '\n'.join(
                save_itl_eps(
                    None,
                    *events,
                    ref=args.relative_to,
                    overlap=args.overlap,
                    header=args.header,
                )
            )
            + '\n'
        )



"""Majis ITL in JSON format sub-module."""

import json
from pathlib import Path

from planetary_coverage.events import EventWindow

from .mapping import JSON_MAPPING
from .schema import MAJIS_ITL_SCHEMAS


def read_itl_json(fname: str | Path) -> list[EventWindow]:
    """Read ITL file in JSON format.

    ITL are validated against Juice SOC and Majis schema.

    """
    data = json.loads(Path(fname).read_text())

    for schema in MAJIS_ITL_SCHEMAS:
        schema.validate(data)

    # Check filename match the one in the file
    filename = Path(data['header']['filename'])
    if not filename or filename.name != fname.name:
        raise FileNotFoundError(filename)

    # Set default values
    defaults = {
        'ITL': Path(filename),
        'COMMENTS': '',
    }

    return [
        EventWindow(
            obs['name'],
            _parse_json(obs, defaults=defaults),
        )
        for obs in data['timeline']
        # Filter only Majis observations
        if obs['instrument'] == 'MAJIS'
    ]


def _parse_json(obs: dict, defaults: dict | None = None) -> dict:
    """Map ITL JSON to timeline keys."""
    # Extract parameters
    params = obs.pop('parameters')

    # Convert JSON keys to EPS keys
    content = {
        JSON_MAPPING.get(key, key.upper()): value for key, value in (obs | params).items()
    }

    # Add extra keys
    content['PRIME'] = content.get('OBSERVATION_TYPE') == 'PRIME'

    # Add default keys (if not present)
    for key, value in defaults.items():
        if key not in content:
            content[key] = value

    # Put the comments at the end (always present from defaults)
    content['COMMENTS'] = content.pop('COMMENTS')

    return content



"""ITL CSV export module."""

from pathlib import Path

from planetary_coverage.events import Event, EventsDict, EventsList, EventWindow

from ...misc.events import flatten_events
from ...misc.export import save_file
from ...misc.time import fmt_datetime


@flatten_events
def save_csv(
    fout: str | Path | None,
    *events: EventsList | EventsDict,
    ref: str | None = None,
    overlap: bool = False,
    sep: str = ';',
) -> Path:
    """Save ITL events to CSV.

    Note
    ----
    By default, ITL blocks must not overlap each other.
    This can be disable with `overlap=True`.

    """
    content = fmt_csv(events, ref=ref, sep=sep)

    return save_file(fout, content, suffix='.csv')


CSV_TIME_REL_KEYS = {
    't_start': ['OBS_START', 'OBS_START_REL'],
    't_end': ['OBS_END', 'OBS_END_REL'],
}


def get_header(
    blocks: list[list | dict] | EventsList | EventsDict,
    ref: str | dict | None = None,
) -> list:
    """Get CSV header.

    The keys are unique and ordered with the following rules:

    - If `OBS_NAME` is present, it will be placed first.

    - `t_start` and `t_end` are required and replaced by `OBS_START`
        and `OBS_END` and placed after `OBS_NAME` (if present).

    - if `ref` time is provided `OBS_START_REL` and `OBS_END_REL`
        are appended after `OBS_START` and `OBS_END`.

    - `COMMENTS` is present, it will always put last.

    """
    keys = dict.fromkeys(key for block in blocks for key in block)

    first, last = [], []
    if 'OBS_NAME' in keys:
        keys.pop('OBS_NAME')
        first.append('OBS_NAME')

    if 't_start' in keys:
        keys.pop('t_start')
        first.append('OBS_START')
    else:
        raise KeyError('Missing `t_start` keyword.')

    if 't_end' in keys:
        keys.pop('t_end')
        first.append('OBS_END')
    else:
        raise KeyError('Missing `t_end` keyword.')

    if ref:
        first.append('OBS_START_REL')
        first.append('OBS_END_REL')

    if 'COMMENTS' in keys:
        keys.pop('COMMENTS')
        last.append('COMMENTS')

    return first + list(keys) + last


def get_value(
    block: dict | Event | EventWindow,
    key: str,
    ref: str | None = None,
) -> str:
    """Format block value.

    This method extract the block value for a given property
    and re-format the datetime if necessary.

    Comments values are escaped with double quotes (`"..."`).

    """
    if key == 'OBS_START':
        return fmt_datetime(block['t_start'])

    if key in ['OBS_START', 'OBS_END']:
        return fmt_datetime(block['t_end'])

    if key == 'OBS_START_REL' and ref:
        return fmt_datetime(block['t_start'], ref=ref)

    if key == 'OBS_END_REL' and ref:
        return fmt_datetime(block['t_end'], ref=ref)

    value = str(block.get(key, ''))

    if key == 'COMMENTS':
        return f'"{value}"'

    return value


def fmt_csv(
    blocks: dict | EventsList | EventsDict,
    ref: str | None = None,
    sep: str = ';',
) -> str:
    """Format CSV format."""
    # Get unique header keys (ordered)
    header = get_header(blocks, ref=ref)

    # Create rows values with separator
    rows = [
        sep.join(get_value(block, key, ref=ref) for key in header) for block in blocks
    ]

    # Merge header and rows
    return ['#' + sep.join(header)] + rows



"""File export miscellaneous module."""

import json
from pathlib import Path


def save_file(
    fout: str | Path | None,
    content: list | dict,
    suffix: str | None = None,
) -> Path | list | dict:
    """Save content to file.

    Raises
    ------
    ValueError
        If the provided extension suffix don't match the one in the filename.

    """
    if fout is None:
        return content

    fout = Path(fout)

    if suffix and fout.suffix != suffix:
        raise ValueError(
            f'Output file name should ends with `{suffix}`: `{fout.name}` provided.'
        )

    if isinstance(content, list):
        _content = '\n'.join(content)
    elif isinstance(content, dict):
        _content = json.dumps(content, indent=2)
    else:
        raise TypeError(
            'Content type must be a `list` or a `dict`, '
            f'not a `{content.__class__.__name__}`'
        )

    fout.write_text(_content, encoding='utf-8')

    return fout



"""Event file miscellaneous module."""

import re
from pathlib import Path

from planetary_coverage import datetime

# EVF key pattern: REF_NAME (COUNT = N)
EVF_KEY = r'(?P<ref>\w+)\s*\(COUNT\s*=\s*(?P<count>\d+)\)'


def ref_count(key: str) -> tuple[str, int]:
    """Extract reference name and count from EVF key."""
    if match := re.match(EVF_KEY, key):
        return match.group('ref').upper(), int(match.group('count'))

    raise KeyError(f'Invalid EVF key pattern: `{key}`. Should be `REF_NAME (COUNT = N)`')


def read_evf(fname: Path | str | list | None, comment: str = '#') -> dict | None:
    """Read and extract EVF references and times."""
    if fname is None:
        return None

    if str(fname).endswith('.evf'):
        lines = Path(fname).read_text().splitlines()
    elif isinstance(fname, list):
        lines = fname
    else:
        lines = fname.splitlines()

    refs = {}
    for line in lines:
        if not line.startswith(comment) and (line_strip := line.strip()):
            dt, key = line_strip.split(maxsplit=1)
            refs[ref_count(key)] = datetime(dt)

    return refs


def ref_key(ref: str | dict | tuple) -> tuple[str, datetime]:
    """Format relative datetime reference key."""
    if isinstance(ref, tuple):
        key, t_ref = ref

    else:
        if not isinstance(ref, dict):
            ref = read_evf(ref)

        if len(ref) > 1:
            raise ValueError(f'Only 1 reference could be provided, not {len(ref)}')

        (ref, count), t_ref = list(ref.items())[0]
        key = f'{ref} (COUNT = {count})'

    return key, t_ref



"""Time miscellaneous module."""

import re

from numpy import datetime64, timedelta64

from planetary_coverage import datetime

from .evf import EVF_KEY, read_evf, ref_key

# Timedelta pattern: 'EVENT_NAME (COUNT = N) [-+][JJJ.]HH:MM:SS[.MS]'
EVF_TD = (
    r'^' + EVF_KEY + r'\s+'
    r'(?P<sign>[+-])?'
    r'(?:(?P<day>\d+)\.)?'
    r'(?P<h>\d{2})'
    r':'
    r'(?P<m>\d{2})'
    r':'
    r'(?P<s>\d{2})'
    r'(?:\.(?P<ms>\d+))?'
)


def get_datetime(time: str, refs: dict | str | list | None = None) -> datetime64:
    """Datetime getter."""
    if match := re.match(EVF_TD, time):
        ref, count, sign, d, h, m, s, ms = match.groups()

        # Extract reference absolute time
        if refs and not isinstance(refs, dict):
            refs = read_evf(refs)

        if not refs or (key := (ref.upper(), int(count))) not in refs:
            raise KeyError(f'Unknown time reference: `{ref} (COUNT = {count})`')

        # Compute time delta to reference
        t_delta = (-1 if sign == '-' else 1) * (
            timedelta64(int(d) if d else 0, 'D')
            + timedelta64(int(h), 'h')
            + timedelta64(int(m), 'm')
            + timedelta64(int(s), 's')
            + (timedelta64(int(1_000 * float(f'0.{ms}')), 'ms') if ms else 0)
        )
        return refs[key] + t_delta

    t, *_ = str(time).strip().split()
    return datetime(t)


def fmt_datetime(
    *times: str | datetime64, ref: str | dict | tuple | None = None
) -> str | list[str]:
    """Format relative datetime for a time or a list of times."""
    if ref is not None:
        key, t_ref = ref_key(ref)

        t = [f'{key}  {fmt_timedelta(datetime(t) - t_ref)}' for t in times]

    else:
        t = [
            datetime(t).item().strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z' for t in times
        ]

    return t if len(times) > 1 else t[0]


def fmt_timedelta(dt: timedelta64) -> str:
    """Format time delta."""
    t = abs(dt.item())
    h, r = divmod(t.seconds, 3600)
    m, s = divmod(r, 60)

    out = '-' if dt < 0 else '+'
    out += f'{t.days:03d}.'
    out += f'{h:02d}:{m:02d}:{s:02d}'
    out += f'.{t.microseconds // 1_000:03d}'

    return out
